/**
* This script provides compatibility functions so we (GraalJS, but also Nashorn, ...) can execute the V8 test suite.
* Note that the functions have been renamed from e.g. %DeoptimizeFunction to V8DeoptimizeFunction, as "%" is no valid part of a function name.
*/

//this method is used e.g. to replace checks like assertTrue(V8HasFastProperties()) that we cannot check properly
function V8AssertIGNORE(a) {
	return undefined;
}
	 
function V8OptimizeFunctionOnNextCall() {
	return undefined;
}

function V8DeoptimizeFunction() {
	return undefined;
}

function V8NeverOptimizeFunction() {
	return undefined;
}

function V8ClearFunctionTypeFeedback() {
	return undefined;
}

function V8RunMicrotasks() {
	return undefined;
}

function V8DebugPrint() {
	for (var i in arguments) {
		print(arguments[i]);
	}
	return undefined;
}

function V8DebugPrintScopes() {
	return undefined;
}

function V8ArgumentsLength(arg) {
	return arg.length;
}

function V8Arguments(arg,i) {
	return arg[i];
}

function V8NotifyContextDisposed() {
	return true;
}

function V8IsConcurrentRecompilationSupported() {
	return true;
}

function V8GetOptimizationStatus() {
	return 0;
}

function V8ToFastProperties(obj) {
	return obj;
}

function V8FlattenString(str) {
	return str;
}

function V8IsValidSmi(value) {
	return Debug.getClass(value).toString() === "class java.lang.Integer";
}

function V8HasFastObjectElements() {
	return true;
}

function V8HasFastSmiOrObjectElements() {
	return true;
}

function V8HasFastSmiElements(array) {
	return Debug.arraytype(array).contains("Int") || Debug.arraytype(array).contains("Byte") || Debug.arraytype(array).contains("Empty") || V8HasFastSmiElementsSlow(array);
}

//helper function
function V8HasFastSmiElementsSlow(array) {
	for (var i=0;i<array.length;i++) {
		if (array.hasOwnProperty(i)) {
			var type = Debug.getClass(array[i]).toString();
			if (!(type === "class java.lang.Integer" || type === "class java.lang.Byte")) {
				return false;
			}
		}
	}
	return true;
}

function V8HasFastDoubleElements(array) {
	return Debug.arraytype(array).contains("Double") || V8HasFastSmiElements(array) || V8HasFastDoubleElementsSlow(array);
}

//helper function
function V8HasFastDoubleElementsSlow(array) {
	for (var i=0;i<array.length;i++) {
		if (array.hasOwnProperty(i)) {
			var type = Debug.getClass(array[i]).toString();
			if (!(type === "class java.lang.Double" || type === "class java.lang.Integer" || type === "class java.lang.Byte")) {
				return false;
			}
		}
	}
	return true;
}

function V8HasFastHoleyElements(obj) {
	return Debug.isHolesArray(obj);
}

function V8HasFastProperties() {
	return true;
}

function V8HasDictionaryElements() {
	return true;
}

function V8HaveSameMap(obj1, obj2) {
	return Debug.haveSameMap(obj1, obj2);
}

function V8HasExternalArrayElements() {
	return true;
}

function V8HasExternalByteElements() {
	return true;
}

function V8HasExternalUnsignedByteElements() {
	return true;
}

function V8HasFixedUint8Elements(ob) {
	return true;
}

function V8HasFixedInt16Elements(ob) {
	return true;
}

function V8HasFixedUint16Elements(ob) {
	return true;
}

function V8HasFixedInt32Elements(ob) {
	return true;
}

function V8HasFixedUint32Elements(ob) {
	return true;
}

function V8HasFixedFloat32Elements(ob) {
	return true;
}

function V8HasFixedFloat64Elements(ob) {
	return true;
}

function V8HasFixedUint8ClampedElements(ob) {
	return true;
}

function gc() {
	return undefined;
}

function V8IsMinusZero(a) {
	if (a === 0 && typeof(a) === "number") {
		return (1/a) === -Infinity;
	} else {
		return false;
	}
}

function V8RunningInSimulator() {
	return false;
}

function V8StringCharCode(st, idx) {
	return st.charAt(idx);
}

function V8StringCharCodeAt(st, idx) {
	return st.charCodeAt(idx);
}

function V8StringFromCharCode(code) {
	return String.fromCharCode(code);
}

function V8UnblockConcurrentRecompilation() {
	return undefined;
}

function V8IsObject(obj) {
	return typeof(obj) === "object" || typeof(obj) === "undefined";
}

function V8GetOptimizationCount(obj) {
	return 0;
}

function V8MaxSmi() {
	return 2147483647;
}

function V8DebugDisassembleFunction() {
	return undefined;
}

function V8SetProperty(obj, name, value) {
	obj[name] = value;
}

function V8SetAllocationTimeout() {
	return undefined;
}

function V8OptimizeObjectForAddingMultipleProperties(obj) {
	return obj;
}

function V8TryMigrateInstance(obj) {
	return obj;
}

function V8StringParseInt(str, rad) {
	return parseInt(str,rad);
}

function externalizeString(str) {
	return str;
}

function V8CallFunction() {
	var thisObj = arguments[0];
	var func = arguments[arguments.length-1];
	var newArgs = [];
	for (var i=1;i<arguments.length-1;i++) {
		newArgs[i-1] = arguments[i];
	}
	return func.call(thisObj,...newArgs);
}

V8Call = V8CallFunction;

function V8Apply() {
	var thisObj = arguments[0];
	var func = arguments[1];
	var newArgs = [];
	for (var i=2;i<arguments.length;i++) {
		newArgs[i-2] = arguments[i];
	}
	return func.apply(thisObj,...newArgs);
}


function V8GlobalParseInt(value, radix) {
	return parseInt(value, radix);
}

function V8SetFlags(flags) {
	return undefined;
}

function V8Break() {
	return undefined;
}

function V8DebugBreak() {
	return undefined;
}

function V8NewString(len, flag) {
	return new String(len);
}

function V8ValueOf(obj) {
	return obj.valueOf();
}

function V8ClassOf(obj) {
	return Debug.className(obj);
}

function V8ObjectFreeze(obj) {
	return obj.freeze();
}

function V8StringCharFromCode(str) {
	return str;
}

function V8SmiLexicographicCompare(a, b) {
	return Debug.stringCompare(a+"",b+"");
}

function V8StringCompare(a, b) {
	return Debug.stringCompare(a+"",b+"");
}

function V8StringCompareRT(a, b) {
	return V8StringCompare(a, b);
}

function V8HasFixedInt8Elements() {
	return false;
}

function V8MakeReferenceError(message) {
	return new ReferenceError(message);
}

function V8AbortJS(message) {
	exit(message);
}

function V8HomeObjectSymbol() {
	return Symbol("__home__");
}

function V8PreventExtensions(obj) {
	return Object.preventExtensions(obj);
}

function V8GetPropertyNames(obj) {
	return Object.keys(obj);
}

function V8NormalizeElements(arr) {
	return arr;
}

function V8SymbolIsPrivate(sym) {
	return false;
}

function V8CreatePrivateSymbol(sym) {
	return Symbol(sym);
}

function V8CreatePrivateOwnSymbol(sym) {
	return Symbol(sym);
}

function V8ArrayBufferNeuter(arr) {
	return arr;
}

function V8GetScript(name) {
	return undefined;
}

function V8AddNamedProperty(object, name, valueParam, prop) {
	var isWritable = !(prop & 1);
	var isEnumerable = !(prop & 2);
	var isConfigurable = !(prop & 4); 
	
	var propDesc = { value: valueParam, writable: isWritable, enumerable: isEnumerable, configurable: isConfigurable };
	
	return Object.defineProperty(object, name, propDesc);
}

function V8SetValueOf(a,b) {
	return a;
}

function V8OneByteSeqStringSetChar() {
}

function V8TwoByteSeqStringSetChar() {
}

function V8IsRegExp(obj) {
	return Object.prototype.toString.call(obj) == "[object RegExp]";
}

function V8IsArray(obj) {
	return Object.prototype.toString.call(obj) == "[object Array]";
}

function V8IsFunction(obj) {
	return Object.prototype.toString.call(obj) == "[object Function]";
}

function V8IsSpecObject(obj) {
	return Object.prototype.toString.call(obj) == "[object Date]";
}

function V8FunctionGetScript(scr) {
}

function V8ConstructDouble(hi,lo) {
	return Debug.constructDouble(hi,lo);
}

function V8DoubleHi(value) {
	return Debug.doubleHi(value);
}

function V8DoubleLo(value) {
	return Debug.doubleLo(value);
}

function V8HasSloppyArgumentsElements(a) {
	return true;
}

function V8MakeError(i,msg) {
	return new Error(msg);
}

function V8AllocateHeapNumber() {
	return 0;
}

function V8DeoptimizeNow() {
	Debug.deoptimize();
}

function V8OptimizeOsr() {
	return undefined;
}

function V8DisassembleFunction(f) {
	Debug.deoptimize(); //not what is expected	
	return undefined;
}

function V8IsStrong(a) {
	return Debug.isStrong(a);
}

var UNDETECTABLE = {};
function V8GetUndetectable() {
	return UNDETECTABLE;
}

function V8FixedArrayGet(vector, slot) {
	return vector[slot];
}

function V8GetTypeFeedbackVector(vector) {
	return undefined;
}

function V8DebugGetLoadedScripts() {
	return undefined;
}

function V8IsSmi(value){
	return typeof(value)==="number" && Math.floor(value) === value && !V8IsMinusZero(value) && Debug.getClass(value).toString() === "class java.lang.Integer" && value != 2147483648;
}

function V8FunctionGetInferredName(a) {
	return a.name;
}

function V8TruncateString(string, index) {
	return string.substring(0,index);
}

function V8MathClz32(a) {
	return Math.clz32(a);
}

function V8ScheduleBreak() {
	return undefined;
}

function V8FormatMessageString(i, m, o, p) {
	if (i < 0) { throw new TypeError("out of bounds"); }
	return new Error(m+o+p);
}

function V8ToMethod(f, o) {
	return f.toMethod(o);
}

function V8RegExpConstructResult() {
	return undefined;
}

function V8DefineAccessorPropertyUnchecked() {
	throw new Error("illegal access");
}

function V8DefineDataPropertyUnchecked() {
	throw new Error("illegal access");
}

function V8NewObjectFromBound(obj) {
	return undefined;
}

function V8IsConstructCall(obj) {
	return false;
}

function V8DeoptimizeNow() {
	return undefined;
}

function V8AtomicsFutexNumWaitersForTesting(a, b) {
	return 0;
}

function V8ToLength(a) {
	return Debug.toLength(a);
}

function V8ToName(a) {
	return Debug.toName(a);
}

function V8ToString(a) {
	return Debug.toStringConv(a);
}

function V8ToPrimitive(a) {
	return Debug.toPrimitive(a);
}

function V8ToPrimitive_Number(a) {
	return Debug.toPrimitiveNumber(a);
}

function V8ToPrimitive_String(a) {
	return Debug.toPrimitiveString(a);
}

function V8ToNumber(a) {
	return Debug.toNumber(a);
}

function V8GetHoleNaNUpper() {
	return undefined;
}

function V8GetHoleNaNLower() {
	return undefined;
}

function V8TailCall() {
	return undefined;
}

